print("import: 'gammapy'")
import gammapy

print("import: 'gammapy.analysis'")
import gammapy.analysis

print("import: 'gammapy.astro'")
import gammapy.astro

print("import: 'gammapy.astro.darkmatter'")
import gammapy.astro.darkmatter

print("import: 'gammapy.astro.population'")
import gammapy.astro.population

print("import: 'gammapy.astro.source'")
import gammapy.astro.source

print("import: 'gammapy.catalog'")
import gammapy.catalog

print("import: 'gammapy.data'")
import gammapy.data

print("import: 'gammapy.irf'")
import gammapy.irf

print("import: 'gammapy.maps'")
import gammapy.maps

print("import: 'gammapy.modeling'")
import gammapy.modeling

print("import: 'gammapy.modeling.models'")
import gammapy.modeling.models

print("import: 'gammapy.scripts'")
import gammapy.scripts

print("import: 'gammapy.stats'")
import gammapy.stats

print("import: 'gammapy.time'")
import gammapy.time

print("import: 'gammapy.utils'")
import gammapy.utils

print("import: 'gammapy.utils.coordinates'")
import gammapy.utils.coordinates

print("import: 'gammapy.utils.random'")
import gammapy.utils.random

