# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Utility functions and classes used throughout Gammapy.

You have to import sub-modules of `gammapy.utils` directly,
the `gammapy.utils` namespace is empty.

Examples::

    from gammapy.utils.interpolation import ScaledRegularGridInterpolator
    from gammapy.utils.nddata import NDDataArray
"""
